/************************************************************************************
DeviceControl RT-Component
Copyright (c) 2010, Kyoto Uni.
All rights reserved.

Contact us if you use this software for sell.
If you use this software not for sell, you can use this software under BSD lisence.
See the files LICENSE.TXT and LICENSE-BSD.TXT for more details.                     
************************************************************************************/

//=============================================================================//
/*!
 *	@file CSerial.cpp
 *	@brief Class for boost based Serial communication class (implementation)
 *  @author kazuyuki kon
 *	@
 */
//=============================================================================//
#include "CSerial.h"

int CSerial::inst = 0;




//=============================================================================//
//Default constructor
//=============================================================================//
CSerial::CSerial()
{

#ifdef WIN32
	m_portName    = "COM1";
#else
	m_portName    = "/dev/ttyUSB0";
#endif
	m_baudrate      = 9600;
	m_characterSize = 8;
	m_parity        = serial_port_base::parity::none;
	m_stopBit       = serial_port_base::stop_bits::one;
	m_flowControl   = serial_port_base::flow_control::none;
	m_thread        = NULL;
	m_port          = NULL;
	m_ioserv        = NULL;

	m_doAsyncRead   = false;
	flushAll();
	doTerm = false;

	inst = inst + 1;
	srand(inst);
	count = rand();

};

//=============================================================================//
//Destruotor
//=============================================================================//
CSerial::~CSerial()
{
}

//=============================================================================//
//open serial port
//=============================================================================//
bool CSerial::openPort()
{
	if( m_port != NULL )delete m_port;
	if( m_ioserv != NULL)delete m_ioserv;
	boost::system::error_code ret;
	m_ioserv = new io_service();
	m_port   = new serial_port( *m_ioserv );
    m_port->open( m_portName.c_str(), ret );

	if( ret.value() != 0 )
    {
		std::cout<< "-------------------------------------------\n"
			     << "Could not open serial port("
		         << m_portName << ", "
				 << m_baudrate << "[bps])" <<"\n"
				 << "Error : " << ret.message() <<"\n"
				 <<"-------------------------------------------"
				 << std::endl;
    	return false;
    }
    else
    {
		//set port settings
		m_port->set_option(serial_port_base::baud_rate(m_baudrate));
		m_port->set_option(serial_port_base::character_size(m_characterSize));

		switch( m_stopBit )
		{
			case serial_port_base::stop_bits::one :
				m_port->set_option(serial_port_base::stop_bits( serial_port_base::stop_bits::one  ));
			break;	
			case serial_port_base::stop_bits::onepointfive :
				m_port->set_option(serial_port_base::stop_bits( serial_port_base::stop_bits::onepointfive ));
			break;	
			case serial_port_base::stop_bits::two:
				m_port->set_option(serial_port_base::stop_bits( serial_port_base::stop_bits::two  ));
			break;	
		}
		
		switch( m_parity )
		{
			case serial_port_base::parity::none:
				m_port->set_option(serial_port_base::parity( serial_port_base::parity::none ));
			break;
			case serial_port_base::parity::even:
				m_port->set_option(serial_port_base::parity( serial_port_base::parity::even ));
			break;
			case serial_port_base::parity::odd:
				m_port->set_option(serial_port_base::parity( serial_port_base::parity::odd ));
			break;

		}
		switch( m_flowControl)
		{
			case serial_port_base::flow_control::none:
				m_port->set_option(serial_port_base::flow_control( serial_port_base::flow_control::none ));
			break;
			case serial_port_base::flow_control::hardware:
				m_port->set_option(serial_port_base::flow_control( serial_port_base::flow_control::hardware ));
			break;
			case serial_port_base::flow_control::software:
				m_port->set_option(serial_port_base::flow_control( serial_port_base::flow_control::software ));
			break;
		}
		//Write status to console
		std::cout<<"--------------------------------------------------\n"
			     << "Connected to ... \n"
				 << "Port        :  " << m_portName    << " \n"
			     << "Baudrate    :  " << m_baudrate    << "[bps]\n" 
				 << "FlowControl :  " << m_flowControl << " \n"
				 << "Stop bit    :  " << m_stopBit     << " \n"
				 << "Parity      :  " << m_parity      << " \n"
				 <<"--------------------------------------------------"
				 << std::endl;

		return m_port->is_open();
    }
}

//=============================================================================//
//Open serial port with async option
//=============================================================================//
bool CSerial::openPort( bool _doAsyncRead )
{
	if( !openPort() )return false;

    if( _doAsyncRead )
	{
		m_doAsyncRead  = _doAsyncRead;

		if( m_doAsyncRead )
			m_port->async_read_some( buffer(m_rbuffer, BUFFERSIZE), bind( &CSerial::readCallback, this, _1,_2) );

		m_thread = new boost::thread( boost::bind( &CSerial::run, this) );
	}

	return true;
}

//=============================================================================//
//Close serial port 
//=============================================================================//
void CSerial::closePort()
{
	doTerm = true;    
	boost::system::error_code ignored;

	m_ioserv->stop();
	m_port->close(ignored);

	if( m_port != NULL)
	{
		delete m_port;
		m_port = NULL;
	}

	if( m_thread != NULL  )
	{	
		delete m_thread;
		m_thread = NULL;
	}

	if( m_ioserv != NULL  )
	{	
		delete m_ioserv;
		m_ioserv = NULL;
	}

	doTerm = false;
	flushAll();
	m_doAsyncRead = false;
};


//=============================================================================//
// Write one character to serial port
//=============================================================================//
size_t CSerial::write(const char _command)
{
	char command[1];
	command[0] = _command;
    return write(command,1);
}

//=============================================================================//
// Write some characters to serial port
// 
//=============================================================================//
size_t CSerial::write(const void* _command, size_t _size)
{
	if( !getConnection() )return 0;

	boost::system::error_code ret;
	size_t sent_size = m_port->write_some( boost::asio::buffer(_command, _size), ret);

	if( ret.value() != 0 )
	{
		std::cout<<ret.message()<<"Error @ CSerial::write()"<<std::endl;
		return 0;
	}

	return sent_size;
}

//=============================================================================//
// Read data from serial com
//=============================================================================//
size_t CSerial::read(char* _command, size_t _size)
{
	if( !getConnection() )return 0;

	boost::system::error_code ret;
    size_t total_size = 0;
	if( m_doAsyncRead )
	{
		//Read data asynchronously
		{
			//boost::mutex::scoped_lock lk(m_mutex);
			m_mutex.lock();
			total_size = std::min(m_readData.size(),_size);
			
			for(int i =0; i< (int)total_size; i++)
			{				
				_command[i] = m_readData.front();
				m_readData.pop_front();
			}	
			m_mutex.unlock();
		}//mutex will be released 
	}
	else
	{
		//Read buffer synchronously
		total_size = m_port->read_some( buffer(_command, _size),ret );

		if( ret.value() != 0 )
		{
			std::cout<<ret.message()<<" @ CSerial::read()"<<std::endl;
			total_size = 0;
		}
	}
	return total_size;
}

//=============================================================================//
// Callback function for read
//=============================================================================//
void CSerial::readCallback( const boost::system::error_code &e, size_t _size )
{
	if( e )
	{
		std::cout<<" Error on read_some() : error code "<< e <<std::endl;
		return;
	}
	
	//Copy data to intermediate buffer
	{
		//boost::mutex::scoped_lock lk(m_mutex);
		m_mutex.lock();

		//Support fot overflow
		if( (getReadDataSize()+_size) > INTERMEDIATE_BUFFER )
		{
			std::cout<<" Warning @ readCallback() : ReadBuffer overflow "<<std::endl;
			int flowSize = (getReadDataSize()+_size) - INTERMEDIATE_BUFFER;
			for(int i = 0;i<flowSize;i++)
				m_readData.pop_front();
		}

		for(int i=0; i<_size; i++)
		{
			m_readData.push_back(m_rbuffer[i]);
			//std::cout<<"ReadCallback "<<m_rbuffer[i]<<" "<<m_readData.size()<<" "<<m_readData.front()<<std::endl; 
		}
		m_mutex.unlock();
	}//mutex will unlocked here.

	//Register read event again.
	//if(!doTerm)
		m_port->async_read_some( buffer(m_rbuffer, BUFFERSIZE), bind( &CSerial::readCallback, this, _1,_2) );

}
